use anyhow::bail;
use bytes::{BufMut, BytesMut};
use prost::Message;
use tokio_util::codec::{Decoder, Encoder};
use crate::MumbleMessage;
use crate::proto::{Authenticate, Version, Ping};

pub struct MumbleTcpCodec {}

impl MumbleTcpCodec {
    pub fn new() -> Self {
        Self {}
    }
}

impl Decoder for MumbleTcpCodec {
    type Item = MumbleMessage;
    type Error = anyhow::Error;

    fn decode(&mut self, src: &mut BytesMut) -> Result<Option<Self::Item>, Self::Error> {
        if src.len() < 6 {
            return Ok(None);
        }

        let (message_type, rest) = src.split_at(2);
        let (message_length, rest) = rest.split_at(4);
        let message_type = u16::from_be_bytes(message_type.try_into()?) as usize;
        let message_length = u32::from_be_bytes(message_length.try_into()?) as usize;
        let mut mumble_message: Option<MumbleMessage> = None;

        if message_length > src.len() {
            src.reserve(message_length+6);
            return Ok(None);
        }

        match message_type {
            0 => {
                //TODO: Version
                let version = Version::decode(rest)?;
                mumble_message = Some(MumbleMessage::Version {data: version});
            }
            3 => {
                let ping = Ping::decode(rest)?;
                mumble_message = Some(MumbleMessage::Ping { data: ping });
            }
            _ => {
                eprintln!("Unknown message type {:?}", message_type);
            }
        }

        let _ = src.split_to(message_length+6);
        src.reserve(6);
        Ok(mumble_message)
    }
}

impl Encoder<MumbleMessage> for MumbleTcpCodec {
    type Error = anyhow::Error;

    fn encode(&mut self, item: MumbleMessage, dst: &mut BytesMut) -> Result<(), Self::Error> {
        let mut message = BytesMut::new();
        match item {
            MumbleMessage::Version { data } => {
                println!("Send Version");
                Version::encode(&data, &mut message)?;
                dst.put_u16(0);
            }
            MumbleMessage::Authenticate { data } => {
                println!("Send Authenticate");
                Authenticate::encode(&data, &mut message)?;
                dst.put_u16(2);
            },
            MumbleMessage::Ping { data } => {
                println!("Send Ping");
                Ping::encode(&data, &mut message)?;
                dst.put_u16(3);
            }
            _ => {
                bail!("Could not encode unknown message")
            }
        }

        dst.put_u32((message.len() as u32) + 6);
        dst.extend_from_slice(&message);
        println!("{dst:?}");
        Ok(())
    }
}